@@ -4,15 +4,19 @@ Main smartglass client
 Common script that handles several subcommands
 See `Commands`
 """
+import os
 import sys
 import logging
 import argparse
 import functools
+import asyncio
+import aioconsole
+import aiohttp
 
+from typing import List
 from logging.handlers import RotatingFileHandler
-from code import InteractiveConsole
 
-from gevent import signal, backdoor
+from xbox.webapi.authentication.models import OAuth2TokenResponse
 
 from xbox.scripts import TOKENS_FILE, CONSOLES_FILE, LOG_FMT, \
     LOG_LEVEL_DEBUG_INCL_PACKETS, VerboseFormatter, ExitCodes
@@ -20,6 +24,7 @@ from xbox.scripts import TOKENS_FILE, CONSOLES_FILE, LOG_FMT, \
 from xbox.handlers import tui, gamepad_input, text_input, fallout4_relay
 from xbox.auxiliary.manager import TitleManager
 
+from xbox.webapi.scripts import CLIENT_ID, CLIENT_SECRET, REDIRECT_URI
 from xbox.webapi.authentication.manager import AuthenticationManager
 from xbox.webapi.common.exceptions import AuthenticationException
 
@@ -27,14 +32,9 @@ from xbox.sg import manager
 from xbox.sg.console import Console
 from xbox.sg.enum import ConnectionState
 
-# REST server imports
-from gevent import pywsgi as rest_pywsgi
-from xbox.rest.app import app as flask_app
-
 
 LOGGER = logging.getLogger(__name__)
 
-REST_DEFAULT_SERVER_PORT = 5557
 REPL_DEFAULT_SERVER_PORT = 5558
 
 
@@ -51,15 +51,16 @@ class Commands(object):
     GamepadInput = 'gamepadinput'
     TextInput = 'textinput'
     TUI = 'tui'
-    RESTServer = 'rest'
 
 
-def parse_arguments(args=None):
+def parse_arguments(args: List[str] = None):
     """
     Parse arguments with argparse.ArgumentParser
 
-    Returns:
-        Namespace: Parsed arguments
+    Args:
+        args: List of arguments from cmdline
+
+    Returns: Parsed arguments
 
     Raises:
         Exception: On generic failure
@@ -84,6 +85,24 @@ def parse_arguments(args=None):
     xbl_token_args.add_argument(
         '--tokens', '-t', type=str, default=TOKENS_FILE,
         help='Tokenfile to load')
+    xbl_token_args.add_argument(
+        "--client-id",
+        "-cid",
+        default=os.environ.get("CLIENT_ID", CLIENT_ID),
+        help="OAuth2 Client ID",
+    )
+    xbl_token_args.add_argument(
+        "--client-secret",
+        "-cs",
+        default=os.environ.get("CLIENT_SECRET", CLIENT_SECRET),
+        help="OAuth2 Client Secret",
+    )
+    xbl_token_args.add_argument(
+        "--redirect-uri",
+        "-ru",
+        default=os.environ.get("REDIRECT_URI", REDIRECT_URI),
+        help="OAuth2 Redirect URI",
+    )
     xbl_token_args.add_argument(
         '--refresh', '-r', action='store_true',
         help="Refresh xbox live tokens in provided token file")
@@ -97,14 +116,6 @@ def parse_arguments(args=None):
         '--liveid', '-l',
         help='LiveID to poweron')
 
-    server_args = argparse.ArgumentParser(add_help=False)
-    server_args.add_argument(
-        '--bind', '-b', default='127.0.0.1',
-        help='Interface address to bind the server')
-    server_args.add_argument(
-        '--port', '-p', type=int, default=0,
-        help='Port to bind to, defaults: (REST: 5557, REPL: 5558)')
-
     """Common argument for interactively choosing console to handle"""
     interactive_arg = argparse.ArgumentParser(add_help=False)
     interactive_arg.add_argument(
@@ -149,12 +160,17 @@ def parse_arguments(args=None):
                  interactive_arg, connection_arg])
 
     """REPL server"""
-    subparsers.add_parser(
+    repl_server_cmd = subparsers.add_parser(
         Commands.REPLServer,
         help='REPL server (interactive console)',
         parents=[logging_args, xbl_token_args,
-                 interactive_arg, connection_arg,
-                 server_args])
+                 interactive_arg, connection_arg])
+    repl_server_cmd.add_argument(
+        '--bind', '-b', default='127.0.0.1',
+        help='Interface address to bind the server')
+    repl_server_cmd.add_argument(
+        '--port', '-p', type=int, default=REPL_DEFAULT_SERVER_PORT,
+        help=f'Port to bind to, default: {REPL_DEFAULT_SERVER_PORT}')
 
     """Fallout relay"""
     subparsers.add_parser(
@@ -186,24 +202,10 @@ def parse_arguments(args=None):
         '--consoles', '-c', default=CONSOLES_FILE,
         help="Previously discovered consoles (json)")
 
-    # FIXME: If possible include startup for REST server here too
-    """
-    REST server
-    NOTE: Only argument parsing is handled in here,
-          The actual start code is in a dedicated script.
-
-          This is required due to gevent monkey patching for
-          the FLASK web-framework to work.
-    """
-    subparsers.add_parser(
-        Commands.RESTServer,
-        help='REST server',
-        parents=[logging_args, xbl_token_args, server_args])
-
     return parser.parse_args(args)
 
 
-def handle_logging_setup(args):
+def handle_logging_setup(args: argparse.Namespace) -> None:
     """
     Determine log level, logfile and special DEBUG_INCL_PACKETS
     via cmdline arguments.
@@ -239,29 +241,45 @@ def handle_logging_setup(args):
         logging.root.addHandler(file_handler)
 
 
-def do_authentication(token_filepath, do_refresh):
+async def do_authentication(args: argparse.Namespace) -> AuthenticationManager:
     """
     Shortcut for doing xbox live authentication (uses xbox-webapi-python lib).
 
     Args:
-        token_filepath (str): Token filepath
-        do_refresh (bool): Whether to refresh tokens
+        args: Parsed arguments
 
-    Returns:
-        AuthenticationManager: An authenticated instance
+    Returns: An authenticated instance of AuthenticationManager
 
     Raises:
         AuthenticationException: If authentication failed
     """
-    auth_mgr = AuthenticationManager.from_file(token_filepath)
-    auth_mgr.authenticate(do_refresh=do_refresh)
-    if do_refresh:
-        auth_mgr.dump(token_filepath)
+    async with aiohttp.ClientSession() as session:
+        auth_mgr = AuthenticationManager(
+            session, args.client_id, args.client_secret, args.redirect_uri
+        )
+
+        # Refresh tokens if we have them
+        if os.path.exists(args.tokens):
+            with open(args.tokens, mode="r") as f:
+                tokens = f.read()
+            auth_mgr.oauth = OAuth2TokenResponse.parse_raw(tokens)
+            await auth_mgr.refresh_tokens()
+
+        # Request new ones if they are not valid
+        if not (auth_mgr.xsts_token and auth_mgr.xsts_token.is_valid()):
+            auth_url = auth_mgr.generate_authorization_url()
+            print(f'Authorize with following URL: {auth_url}\n')
+
+            code = input('Enter received authorization code: ')
+            await auth_mgr.request_tokens(code)
+
+        with open(args.tokens, mode="w") as f:
+            f.write(auth_mgr.oauth.json())
 
     return auth_mgr
 
 
-def choose_console_interactively(console_list):
+def choose_console_interactively(console_list: List[Console]):
     """
     Choose a console to use via user-input
 
@@ -293,13 +311,12 @@ def choose_console_interactively(console_list):
     return console_list[int(response)]
 
 
-def cli_discover_consoles(args):
+async def cli_discover_consoles(args: argparse.Namespace) -> List[Console]:
     """
     Discover consoles
     """
-    LOGGER.info('Sending discovery packets to IP: {0}'
-                .format('IP: ' + args.address if args.address else '<MULTICAST>'))
-    discovered = Console.discover(addr=args.address, timeout=1)
+    LOGGER.info(f'Sending discovery packets to IP: {args.address}')
+    discovered = await Console.discover(addr=args.address, timeout=1)
 
     if not len(discovered):
         LOGGER.error('No consoles discovered')
@@ -320,12 +337,17 @@ def cli_discover_consoles(args):
     return discovered
 
 
-def main(command=None):
+async def main_async(loop: asyncio.AbstractEventLoop, command: Commands = None) -> ExitCodes:
     """
-    Main entrypoint
+    Async Main entrypoint
+
+    Args:
+        command (Commands):
+
+    Returns:
+         None
     """
-    auth_manager = None
-    repl_server_handle = None  # Used for Command.REPLServer
+    auth_manager: AuthenticationManager = None
 
     if command:
         # Take passed command and append actual cmdline
@@ -342,11 +364,7 @@ def main(command=None):
     command = args.command
     LOGGER.debug('Chosen command: {0}'.format(command))
 
-    if command == Commands.RESTServer:
-        LOGGER.info('Make sure you used the dedicated \'xbox-rest-server\' script'
-                    ' to start the REST server!')
-
-    elif 'interactive' in args and args.interactive and \
+    if 'interactive' in args and args.interactive and \
          (args.address or args.liveid):
         LOGGER.error('Flag \'--interactive\' is incompatible with'
                      ' providing an IP address (--address) or LiveID (--liveid) explicitly')
@@ -363,25 +381,21 @@ def main(command=None):
 
     print('Xbox SmartGlass main client started')
 
-    if command == Commands.RESTServer:
+    if 'tokens' in args:
         """
-        REST Server
+        Do Xbox live authentication
         """
+        LOGGER.debug('Command {0} supports authenticated connection'.format(command))
+        print('Authenticating...')
+        try:
+            auth_manager = await do_authentication(args)
+        except AuthenticationException:
+            LOGGER.exception('Authentication failed!')
+            LOGGER.error("Please re-run xbox-authenticate to get a fresh set")
+            sys.exit(ExitCodes.AuthenticationError)
+        print('Authentication done')
 
-        if args.port == 0:
-            LOGGER.info('No defaults provided, '
-                        'Setting REST server port to {0}'.format(REST_DEFAULT_SERVER_PORT))
-            args.port = REST_DEFAULT_SERVER_PORT
-
-        print('Xbox Smartglass REST server started on {0}:{1}'.format(
-            args.bind, args.port
-        ))
-
-        flask_app.token_file = args.tokens
-        server = rest_pywsgi.WSGIServer((args.bind, args.port), flask_app)
-        server.serve_forever()
-        sys.exit(ExitCodes.OK)
-    elif command == Commands.TUI:
+    if command == Commands.TUI:
         """
         Text user interface (powered by urwid)
         """
@@ -391,20 +405,8 @@ def main(command=None):
             LOGGER.debug('Removing StreamHandler {0} from root logger'.format(h))
             logging.root.removeHandler(h)
 
-        sys.exit(tui.run_tui(args.consoles, args.address,
-                             args.liveid, args.tokens, args.refresh))
-
-    elif 'tokens' in args:
-        """
-        Do Xbox live authentication
-        """
-        LOGGER.debug('Command {0} supports authenticated connection'.format(command))
-        try:
-            auth_manager = do_authentication(args.tokens, args.refresh)
-        except AuthenticationException:
-            LOGGER.exception('Authentication failed!')
-            LOGGER.error("Please re-run xbox-authenticate to get a fresh set")
-            sys.exit(ExitCodes.AuthenticationError)
+        await tui.run_tui(loop, args.consoles, auth_manager)
+        return ExitCodes.OK
 
     elif command == Commands.PowerOn:
         """
@@ -416,14 +418,14 @@ def main(command=None):
 
         LOGGER.info('Sending poweron packet for LiveId: {0} to {1}'
                     .format(args.liveid,
-                            'IP: ' + args.address if args.address else '<MULTICAST>'))
-        Console.power_on(args.liveid, args.address, tries=10)
+                            'IP: ' + args.address if args.address else 'MULTICAST'))
+        await Console.power_on(args.liveid, args.address, tries=10)
         sys.exit(0)
 
     """
     Discovery
     """
-    discovered = cli_discover_consoles(args)
+    discovered = await cli_discover_consoles(args)
 
     if command == Commands.Discover:
         """
@@ -441,7 +443,7 @@ def main(command=None):
         """Powering off all discovered consoles"""
         for c in discovered:
             print('Powering off console {0}'.format(c))
-            c.power_off()
+            await c.power_off()
         sys.exit(ExitCodes.OK)
 
     """
@@ -477,15 +479,21 @@ def main(command=None):
     console.on_timeout += \
         lambda x: LOGGER.error('Timeout occured!') or sys.exit(1)
 
-    userhash = auth_manager.userinfo.userhash
-    xtoken = auth_manager.xsts_token
+    userhash = ''
+    xsts_token = ''
+    
+    if auth_manager:
+        userhash = auth_manager.xsts_token.userhash
+        xsts_token = auth_manager.xsts_token.token
 
-    LOGGER.debug('Authentication info:')
-    LOGGER.debug('Userhash: {0}'.format(userhash))
-    LOGGER.debug('XToken: {0}'.format(xtoken))
+        LOGGER.debug('Authentication info:')
+        LOGGER.debug('Userhash: {0}'.format(userhash))
+        LOGGER.debug('XToken: {0}'.format(xsts_token))
+    else:
+        LOGGER.info('Running in anonymous mode')
 
     LOGGER.info('Attempting connection...')
-    state = console.connect(userhash, xtoken.jwt)
+    state = await console.connect(userhash, xsts_token)
     if state != ConnectionState.Connected:
         LOGGER.error('Connection failed! Console: {0}'.format(console))
         sys.exit(1)
@@ -493,19 +501,17 @@ def main(command=None):
     # FIXME: Waiting explicitly
     LOGGER.info('Connected to console: {0}'.format(console))
     LOGGER.debug('Waiting a second before proceeding...')
-    console.wait(1)
+    await console.wait(1)
 
     if command == Commands.PowerOff:
         """
         Power off (single console)
         """
         print('Powering off console {0}'.format(console))
-        console.power_off()
+        await console.power_off()
         sys.exit(ExitCodes.OK)
 
-    elif command == Commands.REPL or \
-            command == Commands.REPLServer:
-
+    elif command == Commands.REPL or command == Commands.REPLServer:
         banner = 'You are connected to the console @ {0}\n'\
                  .format(console.address)
         banner += 'Type in \'console\' to acccess the object\n'
@@ -515,23 +521,17 @@ def main(command=None):
 
         if command == Commands.REPL:
             LOGGER.info('Starting up local REPL console')
-            repl_local = InteractiveConsole(locals=scope_vars)
-            repl_local.interact(banner)
+            console = aioconsole.AsynchronousConsole(locals=scope_vars, loop=loop)
+            await console.interact(banner)
+
         else:
-
-            if args.port == 0:
-                LOGGER.info('No defaults provided, '
-                            'Setting REPL server port to {0}'.format(REPL_DEFAULT_SERVER_PORT))
-                args.port = REPL_DEFAULT_SERVER_PORT
-
             startinfo = 'Starting up REPL server @ {0}:{1}'.format(args.bind, args.port)
             print(startinfo)
             LOGGER.info(startinfo)
 
-            repl_server_handle = backdoor.BackdoorServer(
-                listener=(args.bind, args.port),
-                banner=banner,
-                locals=scope_vars)
+            server = await aioconsole.start_interactive_server(
+                host=args.bind, port=args.port, loop=loop)
+            await server
 
     elif command == Commands.FalloutRelay:
         """
@@ -540,7 +540,9 @@ def main(command=None):
         print('Starting Fallout 4 relay service...')
         console.add_manager(TitleManager)
         console.title.on_connection_info += fallout4_relay.on_connection_info
-        console.start_title_channel(title_id=fallout4_relay.FALLOUT_TITLE_ID)
+        await console.start_title_channel(
+            title_id=fallout4_relay.FALLOUT_TITLE_ID
+        )
         print('Fallout 4 relay started')
     elif command == Commands.GamepadInput:
         """
@@ -548,7 +550,7 @@ def main(command=None):
         """
         print('Starting gamepad input handler...')
         console.add_manager(manager.InputManager)
-        gamepad_input.input_loop(console)
+        await gamepad_input.input_loop(console)
     elif command == Commands.TextInput:
         """
         Text input
@@ -559,19 +561,21 @@ def main(command=None):
         console.text.on_systemtext_input += functools.partial(text_input.on_text_input, console)
         console.text.on_systemtext_done += text_input.on_text_done
 
-    LOGGER.debug('Installing gevent SIGINT handler')
-    signal.signal(signal.SIGINT, lambda *a: console.protocol.stop())
+        while True:
+            try:
+                await asyncio.sleep(1)
+            except KeyboardInterrupt:
+                print('Quitting text input handler')
+                return ExitCodes.OK
 
-    if repl_server_handle:
-        LOGGER.debug('Starting REPL server protocol')
 
-    LOGGER.debug('Starting console.protocol.serve_forever()')
-    console.protocol.serve_forever()
-
-    LOGGER.debug('Protocol serving exited')
-    if repl_server_handle:
-        LOGGER.debug('Stopping REPL server protocol')
-        repl_server_handle.stop()
+def main(command: Commands = None):
+    LOGGER.debug('Entering main_async')
+    try:
+        loop = asyncio.get_event_loop()
+        loop.run_until_complete(main_async(loop, command))
+    except KeyboardInterrupt:
+        pass
 
 
 def main_discover():
